@@ -28,17 +28,17 @@ from elementpath.aliases import Emptiable
 from elementpath.protocols import ElementProtocol
 from elementpath.exceptions import ElementPathError
 from elementpath.tdop import MultiLabel
-from elementpath.helpers import OCCURRENCE_INDICATORS, EQNAME_PATTERN, \
-    XML_NEWLINES_PATTERN, is_xml_codepoint, node_position
+from elementpath.helpers import OCCURRENCE_INDICATORS, Patterns, \
+    is_xml_codepoint, node_position
 from elementpath.namespaces import get_expanded_name, split_expanded_name, \
     XPATH_FUNCTIONS_NAMESPACE, XSD_NAMESPACE
 from elementpath.datatypes import xsd10_atomic_types, NumericProxy, QName, Date10, \
     DateTime10, Time, AnyURI, UntypedAtomic, AtomicType, NumericType, NMToken, \
     Idref, Entity
 from elementpath.sequence_types import is_sequence_type, match_sequence_type
-from elementpath.etree import defuse_xml, etree_iter_paths
-from elementpath.xpath_nodes import XPathNode, ElementNode, TextNode, AttributeNode, \
-    NamespaceNode, DocumentNode, ProcessingInstructionNode, CommentNode
+from elementpath.etree import defuse_xml
+from elementpath.xpath_nodes import XPathNode, ElementNode, NamespaceNode, \
+    DocumentNode, EtreeElementNode, SchemaElementNode
 from elementpath.tree_builders import get_node_tree
 from elementpath.xpath_tokens import XPathToken, ValueToken, XPathFunction, XPathConstructor
 from elementpath.serialization import get_serialization_params, serialize_to_xml, \
@@ -901,7 +901,7 @@ def evaluate_format_date_time_functions(self: XPathFunction, context: ContextTyp
                     except (KeyError, ValueError) as err:
                         raise self.error('FOFD1340', str(err)) from None
 
-                elif EQNAME_PATTERN.search(calendar) is None:
+                elif Patterns.extended_qname.search(calendar) is None:
                     raise self.error('FOFD1340', f'Invalid calendar argument {calendar!r}')
             else:
                 result.append('[' if not result else ', ')
@@ -1073,65 +1073,24 @@ def evaluate_path_function(self: XPathFunction, context: ContextType = None) ->
     elif context is None:
         raise self.missing_context()
 
-    if isinstance(context, XPathSchemaContext):
-        return []
-    elif not self:
+    if not self:
         item = context.item
     else:
         item = self.get_argument(context)
         if item is None:
             return []
 
-    suffix = ''
-    if isinstance(item, DocumentNode):
-        return '/'
-    elif isinstance(item, (ElementNode, CommentNode, ProcessingInstructionNode)):
-        elem = item.elem
-    elif not isinstance(item, XPathNode) or item.parent is None:
+    if not isinstance(item, XPathNode):
         return []
-    elif isinstance(item, TextNode):
-        elem = item.parent.elem
-        suffix = '/text()[1]'
-    elif isinstance(item, AttributeNode) and item.name is not None:
-        elem = item.parent.elem
-        if item.name.startswith('{'):
-            suffix = f'/@Q{item.name}'
-        else:
-            suffix = f'/@{item.name}'
-    elif isinstance(item, NamespaceNode):
-        elem = item.parent.elem
-        if item.prefix:
-            suffix = f'/namespace::{item.prefix}'
-        else:
-            suffix = f'/namespace::*[Q{{{XPATH_FUNCTIONS_NAMESPACE}}}local-name()=""]'
+    elif context.root is None or (root_node := item.root_node) is not context.root:
+        # The context has no root or the root is not the root of the item node
+        return []
+    elif not isinstance(root_node, (DocumentNode, SchemaElementNode)):
+        # It's a fragment: add fn:root() to select the root position
+        path = item.path[len(root_node.path):]
+        return f"Q{{{XPATH_FUNCTIONS_NAMESPACE}}}root(){path}"
     else:
-        return []
-
-    if isinstance(context.root, DocumentNode):
-        root = context.root.getroot().elem
-        if root.tag.startswith('{'):
-            path = f'/Q{root.tag}[1]'
-        else:
-            path = f'/Q{{}}{root.tag}[1]'
-    elif context.root is None:
-        return []
-    else:
-        # If root is an element use the function that returns the root of the tree
-        root = context.root.elem
-        path = 'Q{%s}root()' % XPATH_FUNCTIONS_NAMESPACE
-
-    if isinstance(item, ProcessingInstructionNode):
-        if item.parent is None or isinstance(item.parent, DocumentNode):
-            return f'/processing-instruction({item.name})[{context.position}]'
-    elif isinstance(item, CommentNode):
-        if item.parent is None or isinstance(item.parent, DocumentNode):
-            return f'/comment()[{context.position}]'
-
-    for e, path in etree_iter_paths(root, path):
-        if e is elem:
-            return path + suffix
-    else:
-        return []
+        return item.path or []
 
 
 @method(function('has-children', nargs=(0, 1), sequence_types=('node()?', 'xs:boolean')))
@@ -1153,7 +1112,7 @@ def evaluate_has_children_function(self: XPathFunction, context: ContextType = N
             raise self.error('XPTY0004', 'argument must be a node')
 
     return isinstance(item, DocumentNode) or \
-        isinstance(item, ElementNode) and (len(item.elem) > 0 or item.elem.text is not None)
+        isinstance(item, EtreeElementNode) and (len(item.obj) > 0 or item.obj.text is not None)
 
 
 @method(function('innermost', nargs=1, sequence_types=('node()*', 'node()*')))
@@ -1335,7 +1294,7 @@ def evaluate_unparsed_text_functions(self: XPathFunction, context: ContextType =
     text = text.lstrip('\ufeff')
 
     if self.symbol == 'unparsed-text-lines':
-        lines = XML_NEWLINES_PATTERN.split(text)
+        lines = Patterns.xml_newlines.split(text)
         return lines[:-1] if lines[-1] == '' else lines
 
     return text
@@ -1492,7 +1451,7 @@ def evaluate_parse_xml_fragment_function(self: XPathFunction, context: ContextTy
         else:
             root = etree.XML(arg)
     except etree.ParseError as err:
-        # A not parsable fragment try to parse including XML data in a dummy element.
+        # A not parsable fragment: try to parse including XML data in a dummy element.
         try:
             dummy_element_node = get_node_tree(
                 root=etree.XML(f'<document>{arg}</document>'),
@@ -1502,7 +1461,7 @@ def evaluate_parse_xml_fragment_function(self: XPathFunction, context: ContextTy
             raise self.error('FODC0006', str(err)) from None
         else:
             assert isinstance(dummy_element_node, ElementNode)
-            return DocumentNode.from_element_node(dummy_element_node)
+            return dummy_element_node.get_document_node(replace=True)
     else:
         return cast(DocumentNode, get_node_tree(
             root=etree.ElementTree(root),
@@ -1592,7 +1551,7 @@ def evaluate_function_name_function(self: XPathFunction, context: ContextType =
     if not isinstance(func, XPathFunction):
         raise self.error('XPTY0004', "argument is not a function")
     else:
-        name = func.name
+        name = func.qname
         return [] if name is None else name
 
 
@@ -1748,17 +1707,15 @@ XPath30Parser.unregister('round')
 @method(function('data', nargs=(0, 1), sequence_types=('item()*', 'xs:anyAtomicType*')))
 def select_data_function(self: XPathFunction, context: ContextType = None) \
         -> Iterator[AtomicType]:
-    if not self:
-        item = self.get_argument(context, default_to_context=True)
-        value = self.data_value(item)
-        if value is None:
-            raise self.error('FOTY0012', "argument node does not have a typed value")
-        elif isinstance(value, list):
-            yield from value
-        else:
-            yield value
-    else:
+    if self.context is not None:
+        context = self.context
+
+    if self:
         yield from self[0].atomization(context)
+    elif context is None:
+        raise self.missing_context()
+    else:
+        yield from self.atomize_item(context.item)
 
 
 @method(function('document-uri', nargs=(0, 1), sequence_types=('node()?', 'xs:anyURI?')))
@@ -1788,12 +1745,12 @@ def evaluate_nilled_function(self: XPathFunction, context: ContextType = None) \
     arg: Optional[NumericType] = self.get_argument(self.context or context, default_to_context=True)
     if arg is None:
         return []
-    elif not isinstance(arg, XPathNode):
+    elif isinstance(arg, XPathNode):
+        result = arg.nilled
+        return result if result is not None else []
+    else:
         raise self.error('XPTY0004', 'an XPath node required')
 
-    result = arg.nilled
-    return result if result is not None else []
-
 
 @method(function('node-name', nargs=(0, 1), sequence_types=('node()?', 'xs:QName?')))
 def evaluate_node_name_function(self: XPathFunction, context: ContextType = None) \
@@ -1802,24 +1759,24 @@ def evaluate_node_name_function(self: XPathFunction, context: ContextType = None
     arg = self.get_argument(self.context or context, default_to_context=True)
     if arg is None:
         return []
-    elif not isinstance(arg, XPathNode):
-        raise self.error('XPTY0004', 'an XPath node required')
-
-    name = arg.name
-    if name is None:
-        return []
-    elif name.startswith('{'):
-        # name is a QName in extended format
-        namespace, local_name = split_expanded_name(name)
-        for pfx, uri in self.parser.namespaces.items():
-            if uri == namespace:
-                if not pfx:
-                    return QName(uri, local_name)
-                return QName(uri, '{}:{}'.format(pfx, local_name))
-        raise self.error('FONS0004', 'no prefix found for namespace {}'.format(namespace))
+    elif isinstance(arg, XPathNode):
+        name = arg.name
+        if name is None:
+            return []
+        elif name.startswith('{'):
+            # name is a QName in extended format
+            namespace, local_name = split_expanded_name(name)
+            for pfx, uri in self.parser.namespaces.items():
+                if uri == namespace:
+                    if not pfx:
+                        return QName(uri, local_name)
+                    return QName(uri, '{}:{}'.format(pfx, local_name))
+            raise self.error('FONS0004', 'no prefix found for namespace {}'.format(namespace))
+        else:
+            # name is a local name
+            return QName(self.parser.namespaces.get('', ''), name)
     else:
-        # name is a local name
-        return QName(self.parser.namespaces.get('', ''), name)
+        raise self.error('XPTY0004', 'an XPath node required')
 
 
 @method(function('string-join', nargs=(1, 2),
@@ -1955,6 +1912,9 @@ def cast_error_type(self: XPathConstructor, value: AtomicType) -> Emptiable[None
 @method('error')
 def nud_error_type_and_function(self: XPathConstructor) -> XPathConstructor:
     self.clear()
+    if not self.parser.parse_arguments:
+        return self
+
     try:
         self.parser.advance('(')
         if self.namespace == XSD_NAMESPACE:
